#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <SDL/SDL.h>
#include <math.h>
#include <signal.h>
#include <unistd.h>
#include <poll.h>
#include <fcntl.h>
#include <errno.h>
#include "a_oscill.h"

#define RBUF_SIZE	65535

FILE *fp = NULL;
struct pollfd fds[2];
int fd, fd_pipe;
int is_pipe;
pid_t cpid = -1;

void a_oscill_set_data(struct a_oscill *oscill, const char *ptr);
void a_oscill_help(struct a_oscill *oscill, const char *ptr);

struct a_oscill_command_list {
	char *namecmd;
	void (*func)(struct a_oscill *oscill, const char *data);
} command_list[] = {
	{ "get time",          a_oscill_get_time_cmd },
	{ "set time",          a_oscill_set_time_cmd },
	{ "get voltage",       a_oscill_get_voltage_cmd },
	{ "set voltage",       a_oscill_set_voltage_cmd },
	{ "get input_time",    a_oscill_get_i_time_cmd },
	{ "set input_time",    a_oscill_set_i_time_cmd },
	{ "get screen_color",  a_oscill_get_scolor_cmd },
	{ "set screen_color",  a_oscill_set_scolor_cmd },
	{ "get curve_color",   a_oscill_get_ccolor_cmd },
	{ "set curve_color",   a_oscill_set_ccolor_cmd },
	{ "get update_screen", a_oscill_get_t_update_screen_cmd },
	{ "set update_screen", a_oscill_set_t_update_screen_cmd },
	{ "set data",	       a_oscill_set_data },
	{ "help",              a_oscill_help },
	{ NULL, NULL }
};

void
a_oscill_help(struct a_oscill *oscill, const char *ptr)
{
	printf("Usage:\n"
	       "    set time [time]          - set new time division (type of time is double)\n"
	       "    get time                 - get value time division\n"
	       "    set voltage [voltage]    - set new voltage division\n"
	       "    get voltage              - get value voltage division\n"
	       "    set curve_color [color]  - set new color of curve\n"
	       "                          color = red or yellow or blue or green\n"
	       "    get curve_color          - get color of curve\n"
	       "    set screen_color [color] - set new color of screen\n"
	       "                           color = red or yellow or blue or green\n"
	       "    get screen_color         - get color of screen\n"
	       "    set input_time [time]    - set new input time division(type of time is double)\n"
	       "    get input_time           - get value input time division\n"
	       "    set update_screen [time] - set new time update screen(type double)\n"
	       "    get update_screen        - get time update screen\n"
	       "    set data [file_name]     - read data in file or pipe and draw in screen\n"
	       "                               if you wish open pipe then first symbol in file_name must be '|'\n"
	       "                               in regime \"set data\" command help not work\n"
	);
}

static void
close_pipe()
{
	int res, status;

	assert(cpid >= 0);
	assert(fd_pipe >= 0);

	close(fd_pipe);

	for (;;) {

		kill(cpid, SIGTERM); 

		do {
			res = waitpid(cpid, &status, 0);
		} while (res == -1 && errno == EINTR);
	
		if (res == -1 && errno != ECHILD)
			fprintf(stderr, "waitpid: %s\n", strerror(errno));

		if (res == cpid || (res == -1 && errno == ECHILD)) {
			cpid = -1;
			break;
		}
	}

	cpid = -1;
	fd_pipe = -1;
	is_pipe = 0;
}

void
a_oscill_set_data(struct a_oscill *oscill, const char *ptr)
{
	FILE *nfp;
	int pipefd[2];
	int pflag;
	int null, res;
	pid_t pid;

	if (ptr == NULL)
		return;

	while (isspace(*ptr))
		++ptr;

	pflag = 0;

	if (*ptr == '|') {
		pflag = 1;

		null = open("/dev/null", O_RDWR);
		if (null == -1) {
			fprintf(stderr, "can't open /dev/null: %s\n",
				strerror(errno));
			return;
		}

		res = pipe(pipefd);
		if (res == -1) {
			fprintf(stderr, "can't create pipe: %s\n",
				strerror(errno));
			close(null);
			return;
		}

		pid = fork();

		switch (pid) {
		case -1:
			fprintf(stderr, "can't create child process: %s\n",
				strerror(errno));
			close(pipefd[0]);
			close(pipefd[1]);
			close(null);
			return;
		case 0:
			signal(SIGPIPE, SIG_DFL);
			signal(SIGCHLD, SIG_DFL);
			close(0);
			close(1);
			close(2);
			dup(null);	/* stdin */
			dup(pipefd[1]); /* stdout */
			dup(null);	/* stderr */
			close(null);
			close(pipefd[0]);
			close(pipefd[1]);
			execl("/bin/sh", "/bin/sh", "-c", ++ptr, NULL);
			_exit(127);
		default:
			break;
		}

		close(null);
		close(pipefd[1]);
	} else {
		nfp = fopen(ptr, "r");

		if (nfp == NULL) {
			fprintf(stderr, "can't open file `%s'\n", ptr);
			return;
		}
	}
	
	if (fp != NULL || is_pipe) {
		fcntl(fd_pipe, F_SETFD, fcntl(fd_pipe, F_GETFD) & ~O_NONBLOCK);
		
		if (is_pipe) {
			close_pipe();
		} else {
			fclose(fp);
		}
	}

	is_pipe = pflag;

	if (is_pipe) {
		assert(pid >= 0);
		cpid = pid;
		fd_pipe = pipefd[0];
		fp = NULL;
	} else {
		fd_pipe = fileno(nfp);
		fp = nfp;
	}

	fcntl(fd_pipe, F_SETFD, fcntl(fd_pipe, F_GETFD) | O_NONBLOCK);
	fds[1].fd = fd_pipe;
	fds[1].events = POLLIN;
	fds[1].revents = 0;
}

int main()
{
	char buff[A_OSCILL_BUFF_SIZE];
	char *ptr;
	struct a_oscill *oscill;
	struct a_oscill_command_list *cmdl;
	char *rb, *tail, *head;
	size_t rbsize;
	int prompt, res;
	double data;

	fd = fileno(stdin);
	fcntl(fd, F_SETFD, fcntl(fd, F_GETFD) | O_NONBLOCK);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGCHLD, SIG_IGN);

	rbsize = RBUF_SIZE;
	rb = malloc(rbsize);
	if (rb == NULL) {
		fprintf(stderr, "can't allocate rb");
		exit(1);
	}
	memset(rb, 0, rbsize);
	tail = head = rb;

	fds[0].fd     = fd;
	fds[0].events = POLLIN;

	fds[1].fd     = -1;
	fds[1].events = 0;

	printf("oscillograph 1.0\n Type 'help' or 'exit' to quit.\n");
	oscill = a_oscill_new(yellow, red);

	a_oscill_set_i_time_div(oscill, 0.025);
	a_oscill_set_time_div(oscill, 0.025);
	a_oscill_set_t_update_screen(oscill, 500);

	prompt = 1;

	for (;;) {

		if (prompt > 0) {
			printf("oscill> ");
			fflush(stdout);
			prompt = 0;
		}
		
		res = poll(fds, 2, -1);

		if (res == -1 && errno != EINTR)
			fprintf(stderr, "poll: %s\n", strerror(errno));

		if (fds[0].revents & POLLIN) {
			
			res = read(fds[0].fd, buff, A_OSCILL_BUFF_SIZE);

			if (res == -1 && (errno == EWOULDBLOCK || errno == EAGAIN))
				goto read_pipe;

			if (res < 0) {
				fprintf(stderr, "read: %s\n", strerror(errno));
				goto read_pipe;
			}

			if (res > 0)
				buff[res-1] = 0;

			if (res == 0 || strncmp("exit", buff, strlen("exit")) == 0)
				goto end_loop;

			cmdl = command_list;
			while (cmdl->namecmd != NULL) {
				if (strncmp(cmdl->namecmd, buff, strlen(cmdl->namecmd)) == 0) {
					ptr = buff + strlen(cmdl->namecmd);
					(*cmdl->func)(oscill, ptr);
				}
				cmdl++;
			}

			prompt = 1;
		}

read_pipe:
		if (fds[1].revents & POLLHUP) {
			fprintf(stderr, "data channel closed\n");
			prompt = 1;
			if (is_pipe)
				close_pipe();
			else
				fclose(fp);
			fp = NULL;
			fd_pipe = -1;
			fds[1].revents = 0;
			fds[1].events = 0;
			fds[1].fd = -1;
		}

		if (fds[1].revents & POLLIN) {
			char *s;
			int space;

			space = rbsize - (head - rb);

			if (space <= 1) {
				head = tail = rb;
				space = rbsize;
			}

			res = read(fd_pipe, head, space-1);
			
			if (res == 0 || 
			   (res == -1 && (errno != EAGAIN &&
					  errno != EWOULDBLOCK))) {
				continue;
			}

			if (res > 0) {
				head += res;
				*head = '\0';
				while (tail <= head) {
					s = strchr(tail, '\n');
					if (s == NULL)
						break;
					data = strtod(tail, NULL);
					a_oscill_linear_approx(oscill, data);
					tail = ++s;
				}
			}
		}
	}

end_loop:

	if (is_pipe)
		close_pipe();

	oscill->a_oscill_deinit(&oscill);

	return 0;
}

